// $Header: Frag.cc,v 1.8 99/07/21 16:51:30 vern Exp $
//
// Copyright (c) 1996-1999
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#include "util.h"
#include "Hash.h"
#include "Frag.h"
#include "NetVar.h"
#include "Sessions.h"

#define MIN_ACCEPTABLE_FRAG_SIZE 64
#define MAX_ACCEPTABLE_FRAG_SIZE 64000

void FragTimer::Init(FragReassembler* arg_f, frag_timer_func arg_timer, int arg_do_expire)
        {
        f = arg_f;
        timer = arg_timer;
        do_expire = arg_do_expire;
        Ref(f);
        }

FragTimer::~FragTimer()
        {
        if ( active )
                {
                f->RemoveTimer(this);
                Unref(f);
                }
        }

void FragTimer::Dispatch(double t, int is_expire)
        {
        if ( ! active )
                {
                f = 0;
                return;
                }

        if ( is_expire && ! do_expire )
                return;

        (f->*timer)(t);

        if ( ! active )
                // The timer invocation deactivated us.  Unref the
                // Reassembler here, as we're not going to get another
                // chance.

                Unref(f);
        }

FragReassembler::FragReassembler(NetSessions* arg_s,
			const struct ip* ip, const u_char* pkt,
			uint32 frag_field, HashKey* k,double t)
: Reassembler(0)
	{
	s = arg_s;
	key = k;
	proto_hdr_len = ip->ip_hl * 4;
	proto_hdr = (struct ip*) new u_char[64];	// max IP header
	// Don't do a structure copy - need to pick up options, too.
	memcpy((void*) proto_hdr, (const void*) ip, proto_hdr_len);

	reassembled_pkt = 0;
	frag_size = 0;	// flag meaning "not known"

	AddFragment(t, ip, pkt, frag_field);

	//Click the timer
	ADD_FRAG_TIMER(&FragReassembler::ExpireTimer, t + frag_timeout, 0);
	
	}

FragReassembler::~FragReassembler()
	{
	delete proto_hdr;
	delete reassembled_pkt;
	delete key;

	is_active = 0;
	}

void FragReassembler::AddFragment(double t, const struct ip* ip, const u_char* pkt,
				uint32 frag_field)
	{
	if ( ip->ip_p != proto_hdr->ip_p || ip->ip_hl != proto_hdr->ip_hl )
		// || ip->ip_tos != proto_hdr->ip_tos
		// don't check TOS, there's at least one stack that actually
		// uses different values, and it's hard to see an associated
		// attack.
		s->Weird("fragment_protocol_inconsistency", ip);

	if ( frag_field & 0x4000 )
		s->Weird("fragment_with_DF", ip);

	int offset = (ntohs(ip->ip_off) & 0x1fff) * 8;
	int len = ntohs(ip->ip_len);
	int hdr_len = proto_hdr->ip_hl * 4;
	int upper_seq = offset + len - hdr_len;

	//Set is_active to 1
	is_active = 1;

	if ( (frag_field & 0x2000) == 0 )
		{
		// Last fragment.
		if ( frag_size == 0 )
			frag_size = upper_seq;
		else if ( upper_seq != frag_size )
			s->Weird("fragment_size_inconsistency", ip);
		}

	else if ( len < MIN_ACCEPTABLE_FRAG_SIZE )
		s->Weird("excessively_small_fragment", ip);

	if ( upper_seq > MAX_ACCEPTABLE_FRAG_SIZE )
		s->Weird("excessively_large_fragment", ip);

	// Do we need to check for consistent options?  That's tricky
	// for things like LSRR that get modified in route.

	// Remove header.
	pkt += hdr_len;
	len -= hdr_len;

	NewBlock(network_time, offset, len, pkt);

	}

void FragReassembler::Overlap(const u_char* b1, const u_char* b2, int n)
	{
	if ( memcmp((const void*) b1, (const void*) b2, n) )
		s->Weird("fragment_inconsistency", proto_hdr);
	else
		s->Weird("fragment_overlap", proto_hdr);
	}

void FragReassembler::BlockInserted(double /* t */, DataBlock* /* start_block */)
	{
	if ( blocks->seq > 0 || ! frag_size )
		// For sure don't have it all yet.
		return;

	// We might have it all - look for contiguous all the way.
	DataBlock* b;
	for ( b = blocks; b->next; b = b->next )
		if ( b->upper != b->next->seq )
			break;

	if ( b->next )
		{
		// We have a hole.
		if ( b->seq >= frag_size )
			s->Weird("fragment_size_inconsistency", proto_hdr);
		else
			return;
		}

	else if ( last_block->upper > frag_size )
		s->Weird("fragment_size_inconsistency", proto_hdr);

	else if ( last_block->upper < frag_size )
		// Missing the tail.
		return;

	u_char* pkt = new u_char[proto_hdr_len + frag_size];
	memcpy((void*) pkt, (const void*) proto_hdr, proto_hdr_len);

	reassembled_pkt = (struct ip*) pkt;
	reassembled_pkt->ip_len = htons(frag_size + proto_hdr_len);
	pkt += proto_hdr_len;

	for ( b = blocks; b; b = b->next )
		memcpy((void*) &pkt[b->seq], (const void*) b->block,
			b->upper - b->seq);

	//Got all fragments.
	//Cancel timers.
	CancelTimers(0);
	
	}

void FragReassembler::ExpireTimer(double t)
        {
	if( !is_active)
		return;

	//Remove all the blocks allotted
	while ( blocks )
        	{
	        DataBlock* b = blocks->next;
       		delete blocks;
        	blocks = b;
        	}

	CancelTimers(1);
	sessions->Remove(this);
	is_active = 0;

	return;
	}

void FragReassembler::AddTimer(frag_timer_func timer, double delta_t, int do_expire)
        {
        Timer* frag_timer = new FragTimer(this, timer, delta_t, do_expire);
        timer_mgr->Add(frag_timer);
        timers.append(frag_timer);
        }

void FragReassembler::DeleteTimer()
        {
        CancelTimers(0);
        }

void FragReassembler::CancelTimers(int callee_is_timer)
        {
        if ( callee_is_timer )
                // Protect ourselves from being completely Unref()'d.
                // The ConnectionTimer will Unref() us after dispatching.
                Ref(this);
                       
        loop_over_list(timers, i)
                {
                timers[i]->Cancel();
                
                // Unref ourselves for this timer, since it won't do it
                // for us now.
                Unref(this);
                }
                
        timers.clear();
        }

