// $Header: /home/ram/u0/vern/bro/RCS/Sessions.cc,v 1.48 2001/09/05 04:24:59 vern Exp $
//
// Copyright (c) 1996-1999
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#include <stdlib.h>
#include <unistd.h>

#include "Net.h"
#include "Event.h"
#include "Timer.h"
#include "NetVar.h"
#include "Sessions.h"

#include "FTP.h"
#include "Finger.h"
#include "Ident.h"
#include "HTTP.h"
#include "Portmap.h"
#include "Telnet.h"
#include "Rlogin.h"
#include "DNS.h"
#include "NTP.h"
#include "ICMP.h"

#include "SteppingStone.h"
#include "BackDoor.h"
#include "InterConn.h"
#include "Discard.h"

NetSessions* sessions;


class NetworkTimer : public Timer {
public:
	NetworkTimer(NetSessions* arg_sess, double arg_t) : Timer(arg_t)
		{ sess = arg_sess; }

	void Dispatch(double t, int is_expire);

protected:
	NetSessions* sess;
};

void NetworkTimer::Dispatch(double t, int is_expire)
	{
	if ( is_expire )
		return;

	sess->HeartBeat(t);
	}


NetSessions::NetSessions()
	{
	TypeList* t = new TypeList();
	t->Append(base_type(TYPE_COUNT));	// source IP address
	t->Append(base_type(TYPE_COUNT));	// dest IP address
	t->Append(base_type(TYPE_COUNT));	// source and dest ports

	ch = new CompositeHash(t);

	Unref(t);

	tcp_conns.SetDeleteFunc(bro_obj_delete_func);
	udp_conns.SetDeleteFunc(bro_obj_delete_func);
	fragments.SetDeleteFunc(bro_obj_delete_func);

	if ( reading_live && net_stats_update )
		timer_mgr->Add(new NetworkTimer(this, 1.0));

	if ( stp_correlate_pair )
		stp_manager = new SteppingStoneManager();
	else
		stp_manager = 0;

	discarder = new Discarder();
	if ( ! discarder->IsActive() )
		{
		delete discarder;
		discarder = 0;
		}
	}

NetSessions::~NetSessions()
	{
	delete ch;
	}

void NetSessions::Done()
	{
	delete stp_manager;
	delete discarder;
	}

void NetSessions::NextPacket(double t, const struct pcap_pkthdr* hdr,
				const u_char* const pkt, int hdr_size)
	{
	uint32 caplen = hdr->caplen - hdr_size;
	if ( caplen < sizeof(struct ip) )
		{
		Weird("truncated_IP", hdr, pkt);
		return;
		}

	const struct ip* ip = (const struct ip *)(pkt + hdr_size);

	uint32 len = ntohs(ip->ip_len);
	if ( hdr->len < len + hdr_size )
		{
		Weird("truncated_IP", hdr, pkt);
		return;
		}

	int ip_hdr_len = ip->ip_hl * 4;
	if ( ones_complement_checksum((void*) ip, ip_hdr_len, 0) != 0xffff )
		{
		Weird("bad_IP_checksum", hdr, pkt);
		return;
		}

	if ( discarder && discarder->NextPacket(ip, len, caplen) )
		return;

	int proto = ip->ip_p;
	if ( proto != IPPROTO_TCP && proto != IPPROTO_UDP &&
	     proto != IPPROTO_ICMP )
		{
		DumpPacket(hdr, pkt);
		return;
		}

	FragReassembler* f = 0;
	uint32 frag_field = ntohs(ip->ip_off);
	if ( (frag_field & 0x3fff) != 0 )
		{
		DumpPacket(hdr, pkt);	// always record fragments

		if ( caplen < len )
			{
			Weird("incompletely_captured_fragment", ip);
			return;	// don't try to reassemble, that's doomed
			}

		f = NextFragment(t, ip, pkt + hdr_size, frag_field);
		ip = f->ReassembledPkt();
		if ( ! ip )
			// It didn't reassemble into anything yet.
			return;


		caplen = len = ntohs(ip->ip_len);
		ip_hdr_len = ip->ip_hl * 4;
		}

	len -= ip_hdr_len;	// remove IP header
	caplen -= ip_hdr_len;

	uint32 min_hdr_len = (proto == IPPROTO_TCP) ?  sizeof(struct tcphdr) :
		(proto == IPPROTO_UDP ? sizeof(struct udphdr) : ICMP_MINLEN);

	if ( len < min_hdr_len )
		{
		Weird("truncated_header", hdr, pkt);
		if ( f )
			Remove(f);	// ###
		return;
		}
	if ( caplen < min_hdr_len )
		{
		Weird("internally_truncated_header", hdr, pkt);
		if ( f )
			Remove(f);	// ###
		return;
		}

	// Where the data starts - if this is a protocol we know about,
	// this gets advanced past the transport header.
	const u_char* data = ((u_char *)ip + ip_hdr_len);

	uint32 src_addr = uint32(ip->ip_src.s_addr);
	uint32 dst_addr = uint32(ip->ip_dst.s_addr);
	uint32 src_port, dst_port;
	Dictionary* d;

	switch ( proto ) {
	case IPPROTO_TCP:
		{
		const struct tcphdr* tp = (const struct tcphdr *) data;
		src_port = tp->th_sport;
		dst_port = tp->th_dport;
		d = &tcp_conns;
		break;
		}

	case IPPROTO_UDP:
		{
		const struct udphdr* up = (const struct udphdr *) data;
		src_port = up->uh_sport;
		dst_port = up->uh_dport;
		d = &udp_conns;
		break;
		}

	case IPPROTO_ICMP:
		{
		const struct icmp* icmpp = (const struct icmp *) data;

		src_port = icmpp->icmp_type;
		dst_port = 0;

		d = &icmp_conns;
		break;
		}
	}

	ListVal* key = BuildConnKey(src_addr, dst_addr, src_port, dst_port);
	HashKey* h = ch->ComputeHash(key, 1);
	if ( ! h )
		internal_error("hash computation failed");

	Connection* conn = (Connection*) d->Lookup(h);
	if ( ! conn )
		{
		conn = NewConn(h, t, src_addr, src_port,
				dst_addr, dst_port, data, proto);
		if ( conn )
			d->Insert(h, conn);
		}
	else
		{
		if ( conn->IsReuse(t, data) )
			{
			Remove(conn);
			conn = NewConn(h, t, src_addr, src_port,
					dst_addr, dst_port, data, proto);
			if ( conn )
				d->Insert(h, conn);
			}
		else
			delete h;
		}

	Unref(key);

	if ( ! conn )
		{
		delete h;
		DumpPacket(hdr, pkt);	// save the whole thing
		return;
		}

	int record_packet = 1;	// whether to record the packet at all
	int record_content = 1;	// whether to record its data
	conn->NextPacket(t, src_addr, src_port, ip, len, caplen, data,
				record_packet, record_content);

	if ( packet_contents )
		{
		int hdr_offset = data - ((u_char *)ip + ip_hdr_len);
		int contents_len = min(len, caplen) - hdr_offset;

		if ( contents_len > 0 )
			{
			BroString* cbs = new BroString(data, contents_len, 1);
			Val* contents = new StringVal(cbs);
			conn->Event(packet_contents, contents);
			}
		}

	if ( f )
		{
		// Above we already recorded the fragment in its entirety.
		f->DeleteTimer();
		Remove(f);	// ###
		}

	else if ( record_packet )
		{
		if ( record_content )
			DumpPacket(hdr, pkt);	// save the whole thing
		else
			{
			int hdr_len = data - pkt;
			DumpPacket(hdr, pkt, hdr_len);	// just save the header
			}
		}
	}

FragReassembler* NetSessions::NextFragment(double  t , const struct ip* ip,
					const u_char* pkt, uint32 frag_field)
	{
	uint32 src_addr = uint32(ip->ip_src.s_addr);
	uint32 dst_addr = uint32(ip->ip_dst.s_addr);
	uint32 frag_id = ntohs(ip->ip_id);	// we actually could skip conv.

	ListVal* key = new ListVal(TYPE_ANY);
	key->Append(new Val(src_addr, TYPE_COUNT));
	key->Append(new Val(dst_addr, TYPE_COUNT));
	key->Append(new Val(frag_id, TYPE_COUNT));

	HashKey* h = ch->ComputeHash(key, 1);
	if ( ! h )
		internal_error("hash computation failed");

	FragReassembler* f = fragments.Lookup(h);
	if ( ! f )
		{
		f = new FragReassembler(this, ip, pkt, frag_field, h,t);
		fragments.Insert(h, f);
		Unref(key);
		return f;
		}

	delete h;
	Unref(key);

	f->AddFragment(t,ip, pkt, frag_field);
	return f;
	}

Connection* NetSessions::FindConnection(Val* v)
	{
	if ( v->Type() != conn_id )
		return 0;

	val_list* vl = v->AsRecord();

	uint32 orig_addr = (*vl)[0]->ForceAsUInt();
	uint32 resp_addr = (*vl)[2]->ForceAsUInt();

	PortVal* orig_portv = (*vl)[1]->AsPortVal();
	PortVal* resp_portv = (*vl)[3]->AsPortVal();

	uint32 orig_port = htons((unsigned short) orig_portv->Port());
	uint32 resp_port = htons((unsigned short) resp_portv->Port());

	ListVal* key = BuildConnKey(orig_addr, resp_addr, orig_port, resp_port);
	HashKey* h = ch->ComputeHash(key, 1);
	if ( ! h )
		internal_error("hash computation failed");

	Dictionary* d;
	if ( orig_portv->IsTCP() )
		d = &tcp_conns;
	else
		d = &udp_conns;

	Connection* conn = (Connection*) d->Lookup(h);

	delete h;
	Unref(key);

	return conn;
	}

void NetSessions::Remove(Connection* c)
	{
	HashKey* k = c->Key();
	if ( k )
		{
		// Zero out c's copy of the key, so that if c has been Ref()'d
		// up, we know on a future call to Remove() that it's no
		// longer in the dictionary.
		c->ClearKey();
		if ( c->IsTCP() )
			{
			if ( ! tcp_conns.RemoveEntry(k) )
				internal_error("connection missing");

			((TCP_Connection*) c)->AnalyzerDone();
			}
		else
			{
			if ( ! udp_conns.RemoveEntry(k) )
				internal_error("connection missing");
			}

		Unref(c);
		delete k;
		}
	}

void NetSessions::Remove(FragReassembler* f)
	{
	HashKey* k = f->Key();
	if ( ! k )
		internal_error("fragment block not in dictionary");

	if ( ! fragments.RemoveEntry(k) )
		internal_error("fragment block missing");

	Unref(f);
	}

void NetSessions::Drain()
	{
	IterCookie* cookie = tcp_conns.InitForIteration();
	TCP_Connection* c;

	while ( (c = tcp_conns.NextEntry(cookie)) )
		{
		if ( ! c->BothClosed() )
			c->Event(connection_pending);
		}
	}

void NetSessions::HeartBeat(double t)
	{
	unsigned int recv = 0;
	unsigned int drop = 0;

	loop_over_list(pkt_srcs, i)
		{
		PktSrc* ps = pkt_srcs[i];

		struct pcap_stat stat;
		if ( pcap_stats(ps->PcapHandle(), &stat) < 0 )
			run_time("problem getting packet filter statistics: %s",
				ps->ErrorMsg());
		else
			{
			recv += stat.ps_recv;
			drop += stat.ps_drop;
			}
		}

	val_list* vl = new val_list;

	vl->append(new Val(t, TYPE_TIME));

	RecordVal* ns = new RecordVal(net_stats);
	ns->Assign(0, new Val(recv, TYPE_COUNT));
	ns->Assign(1, new Val(drop, TYPE_COUNT));

	// Turns out that stat.ps_ifdrop isn't even initialized.
	// ns->Assign(2, new Val(stat.ps_ifdrop, TYPE_COUNT));
	ns->Assign(2, new Val(0, TYPE_COUNT));

	vl->append(ns);

	mgr.QueueEvent(net_stats_update, vl);

	timer_mgr->Add(new NetworkTimer(this, t + heartbeat_interval));
	}

void NetSessions::GetStats(SessionStats& s) const
	{
	s.num_TCP_conns = tcp_conns.Length();
	s.num_UDP_conns = udp_conns.Length();
	s.num_ICMP_conns = icmp_conns.Length();
	s.num_fragments = fragments.Length();
	s.num_timers = timer_mgr->Size();

	s.max_TCP_conns = tcp_conns.MaxLength();
	s.max_UDP_conns = udp_conns.MaxLength();
	s.max_ICMP_conns = icmp_conns.MaxLength();
	s.max_fragments = fragments.MaxLength();
	s.max_timers = timer_mgr->PeakSize();
	}

ListVal* NetSessions::BuildConnKey(uint32 src_addr, uint32 dst_addr,
					uint32 src_port, uint32 dst_port)
	{
	ListVal* key = new ListVal(TYPE_ANY);

	// Lookup up connection based on canonical ordering, which is
	// the smaller of <src addr, src port> and <dst addr, dst port>
	// followed by the other.
	if ( src_addr < dst_addr ||
	     (src_addr == dst_addr && src_port < dst_port) )
		{
		key->Append(new Val(src_addr, TYPE_COUNT));
		key->Append(new Val(dst_addr, TYPE_COUNT));
		key->Append(new Val(uint32((src_port << 16) | dst_port), TYPE_COUNT));
		}
	else
		{
		key->Append(new Val(dst_addr, TYPE_COUNT));
		key->Append(new Val(src_addr, TYPE_COUNT));
		key->Append(new Val(uint32((dst_port << 16) | src_port), TYPE_COUNT));
		}

	return key;
	}

Connection* NetSessions::NewConn(HashKey* k, double t,
		uint32 pkt_src_addr, uint32 pkt_src_port,
		uint32 pkt_dst_addr, uint32 pkt_dst_port,
		const u_char* data, int proto)
	{
	switch ( proto ) {
	case IPPROTO_TCP:
		return NewConn(k, t, pkt_src_addr, pkt_src_port,
				pkt_dst_addr, pkt_dst_port,
				(const struct tcphdr *) data);

	case IPPROTO_UDP:
		return NewConn(k, t, pkt_src_addr, pkt_src_port,
				pkt_dst_addr, pkt_dst_port,
				(const struct udphdr *) data);

	case IPPROTO_ICMP:
		return NewConn(k, t, pkt_src_addr, pkt_src_port,
				pkt_dst_addr, pkt_dst_port,
				(const struct icmp *) data);

	default:
		internal_error("bad protocol in NetSessions::NewConn");
		return 0;
	}
	}

TCP_Connection* NetSessions::NewConn(HashKey* k, double t,
			uint32 pkt_src_addr, uint32 pkt_src_port,
			uint32 pkt_dst_addr, uint32 pkt_dst_port,
			const struct tcphdr* tp)
	{
	TCP_Connection* c = 0;

	if ( ! (tp->th_flags & TH_SYN) || (tp->th_flags & TH_ACK) )
		{
		// The new connection is starting either without a SYN,
		// or with a SYN ack.  This means it's a partial connection.

		if ( ! partial_connection_ok )
			return 0;

		uint32 src_port = ntohs(pkt_src_port);
		uint32 dst_port = ntohs(pkt_dst_port);

		if ( tp->th_flags & TH_SYN && ! tcp_SYN_ack_ok )
			return 0;

		// Try to guess true responder by the port numbers.
		// This test should really be for whether the source
		// port is one we interpret, and the destination
		// port isn't.  (Also, we might think that for SYN acks
		// we could safely flip the roles, but that doesn't work
		// for stealth scans.)
		int flip_roles = (src_port < 1024 && dst_port >= 1024);

		if ( flip_roles )
			{
			uint32 t = pkt_src_addr;
			pkt_src_addr = pkt_dst_addr;
			pkt_dst_addr = t;

			t = pkt_src_port;
			pkt_src_port = pkt_dst_port;
			pkt_dst_port = t;
			}
		}

	int is_analyzed = 1;
	int service = ntohs(pkt_dst_port);
	switch ( service ) {
	case 21:
		if ( ftp_request || ftp_reply )
			c = new FTP_Conn(this, k, t,
					pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, tp);
		break;

	case 23:
		if ( login_failure || login_success ||
		     login_input_line || login_output_line )
			c = new TelnetConn(this, k, t,
					pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, tp);
		break;

	case 513:
		if ( login_failure || login_success ||
		     login_input_line || login_output_line )
			c = new RloginConn(this, k, t,
					pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, tp);
		break;

	case 79:
		if ( finger_request || finger_reply )
			c = new FingerConn(this, k, t,
					pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, tp);
		break;

	case 113:
		if ( ident_request || ident_reply || ident_error )
			c = new IdentConn(this, k, t,
					pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, tp);
		break;

	case 80:
	case 8080:
	case 8000:
		if ( http_request || http_reply )
			c = new HTTP_Conn(this, k, t,
					pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, tp);
		break;

	case 53:
		if ( dns_request || dns_full_request )
			c = new TCP_DNS(this, k, t,
					pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, tp);
		break;

	case 111:
		if ( pm_request )
			c = new PortmapperConn(this, k, t,
					pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, tp);
		break;
	}

	if ( ! c )
		{
		if ( new_connection_contents )
			c = new TCP_ConnectionContents(this, k, t,
						pkt_src_addr, pkt_src_port,
						pkt_dst_addr, pkt_dst_port, tp);
		else
			{
			c = new TCP_Connection(this, k, t,
						pkt_src_addr, pkt_src_port,
						pkt_dst_addr, pkt_dst_port, tp);

			is_analyzed = (non_analyzed_lifetime == 0.0);
			}
		}

	c->Init();

	if ( conn_stats )
		c->AddAnalyzer(new TCP_Stats(c));

	if ( stp_correlate_pair &&
	     (service == 22 || service == 23 || service == 513) ) 
		c->AddAnalyzer(new SteppingStoneAnalyzer(c, stp_manager));

	if ( backdoor_stats || rlogin_signature_found || 
	     telnet_signature_found || ssh_signature_found ||
	     root_backdoor_signature_found || ftp_signature_found ||
	     napster_signature_found )
		c->AddAnalyzer(new BackDoorAnalyzer(c));

	if ( interconn_stats )
		c->AddAnalyzer(new InterConnAnalyzer(c));

	if ( new_connection )
		c->Event(new_connection);
	else if ( new_connection_contents )
		c->Event(new_connection_contents);

	if ( ! is_analyzed && ! c->HasAnalyzers() )
		c->SetLifetime(non_analyzed_lifetime);

	return c;
	}

UDP_Connection* NetSessions::NewConn(HashKey* k, double t,
			uint32 pkt_src_addr, uint32 pkt_src_port,
			uint32 pkt_dst_addr, uint32 pkt_dst_port,
			const struct udphdr* up)
	{
	UDP_Connection* c = 0;

	if ( ntohs(pkt_src_port) == 111 )
		{
		// Make a guess that we're seeing the tail half of
		// a portmapper connection.
		uint32 t = pkt_src_addr;
		pkt_src_addr = pkt_dst_addr;
		pkt_dst_addr = t;

		t = pkt_src_port;
		pkt_src_port = pkt_dst_port;
		pkt_dst_port = t;
		}

	switch ( ntohs(pkt_dst_port) ) {
	case 53:
		if ( dns_request || dns_full_request )
			c = new UDP_DNS(this, k, t,
					pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, up);
		break;

	case 111:
		if ( pm_request )
			c = new UDP_Portmapper(this, k, t,
					pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, up);
		break;

	case 123:
		if ( ntp_message )
			c = new NTP_Session(this, k, t,
					pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, up);
		break;
	}

	if ( ! c )
		{
		c = new UDP_Connection(this, k, t, pkt_src_addr, pkt_src_port,
					pkt_dst_addr, pkt_dst_port, up);

		if ( non_analyzed_lifetime != 0.0 )
			c->SetLifetime(non_analyzed_lifetime);
		}

	return c;
	}

ICMP_Connection* NetSessions::NewConn(HashKey* k, double t,
			uint32 pkt_src_addr, uint32 pkt_src_port,
			uint32 pkt_dst_addr, uint32 pkt_dst_port,
			const struct icmp* icmpp)
	{
	switch ( icmpp->icmp_type ) {
	case ICMP_ECHO:
	case ICMP_ECHOREPLY:
		if ( icmp_echo_request || icmp_echo_reply )
			return new ICMP_Echo(this, k, t,
						pkt_src_addr, pkt_src_port,
						pkt_dst_addr, pkt_dst_port,
						icmpp);

	case ICMP_UNREACH:
		if ( icmp_unreachable )
			return new ICMP_Unreach(this, k, t,
						pkt_src_addr, pkt_src_port,
						pkt_dst_addr, pkt_dst_port,
						icmpp);

	default:
		ICMP_Connection* c = new ICMP_Connection(this, k, t,
						pkt_src_addr, pkt_src_port,
						pkt_dst_addr, pkt_dst_port,
						icmpp);

		if ( non_analyzed_lifetime != 0.0 )
			c->SetLifetime(non_analyzed_lifetime);

		return c;
	}
	}

void NetSessions::DumpPacket(const struct pcap_pkthdr* hdr,
				const u_char* pkt, int len)
	{
	if ( ! pkt_dump )
		return;

	if ( len == 0 )
		pcap_dump((u_char*) pkt_dump, hdr, pkt);
	else
		{
		struct pcap_pkthdr h = *hdr;
		h.caplen = len;
		if ( h.caplen > hdr->caplen )
			internal_error("bad modified caplen");
		pcap_dump((u_char*) pkt_dump, &h, pkt);
		}
	}

void NetSessions::Internal(const char* msg, const struct pcap_pkthdr* hdr,
				const u_char* pkt)
	{
	DumpPacket(hdr, pkt);
	internal_error(msg);
	}

void NetSessions::Weird(const char* name,
			const struct pcap_pkthdr* hdr, const u_char* pkt)
	{
	if ( hdr )
		DumpPacket(hdr, pkt);

	if ( net_weird )
		{
		val_list* vl = new val_list;
		vl->append(new StringVal(name));
		mgr.QueueEvent(net_weird, vl);
		}
	else
		fprintf(stderr, "weird: %.06f %s\n", network_time, name);
	}

void NetSessions::Weird(const char* name, const struct ip* ip)
	{
	if ( flow_weird )
		{
		val_list* vl = new val_list;
		vl->append(new StringVal(name));
		vl->append(new AddrVal(ip->ip_src.s_addr));
		vl->append(new AddrVal(ip->ip_dst.s_addr));
		mgr.QueueEvent(flow_weird, vl);
		}
	else
		fprintf(stderr, "weird: %.06f %s\n", network_time, name);
	}
